package fr.amap.lidar.amapvox.commons;

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
import fr.amap.commons.util.io.file.FileManager;
import fr.amap.commons.util.jar.ManifestMetaData;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import javax.vecmath.Point3d;
import javax.vecmath.Point3i;
import org.apache.log4j.Logger;

/**
 *
 * @author calcul
 */
public class VoxelSpaceInfos {

    private final static Logger LOGGER = Logger.getLogger(VoxelSpaceInfos.class);

    private Point3d minCorner;
    private Point3d maxCorner;
    private Point3i split;
    private Type type; //ALS ou TLS
    private final Point3d voxelSize;
    private float resolution;
    private float maxPAD = 5.0f;
    private LeafAngleDistribution.Type ladType = LeafAngleDistribution.Type.SPHERIC;
    private double[] ladParams;
    private String[] columnNames;
    private List<String> columnNamesList;
    private String buildVersion;
    // default decimal format (temporary, should be read from .vox metadata)
    private int nFractionDigits;
    private HashMap<String, String> properties = new HashMap();
    private int nLine;

    public enum Type {
        ALS(1),
        TLS(2);

        private final int type;

        Type(int type) {
            this.type = type;
        }
    }

    public VoxelSpaceInfos() {
        this.voxelSize = new Point3d();
    }

    public VoxelSpaceInfos(Point3d minCorner, Point3d maxCorner, Point3i split) {
        this.minCorner = minCorner;
        this.maxCorner = maxCorner;
        this.split = split;
        this.voxelSize = new Point3d();
        updateVoxelSize();
    }

    public String getProperty(String key) {
        return properties.get(key.toLowerCase());
    }

    public void addProperty(String key, String value) {
        properties.put(key.toLowerCase(), value);
    }

    private void presetProperties() throws IOException {

        try {

            for (String key : properties.keySet()) {
                String value = properties.get(key);
                switch (key) {
                    case "min_corner":
                        String[] minCornerArray = value.split(" ");
                        minCorner = new Point3d(Double.valueOf(minCornerArray[0]),
                                Double.valueOf(minCornerArray[1]),
                                Double.valueOf(minCornerArray[2]));
                        break;
                    case "max_corner":
                        String[] maxCornerArray = value.split(" ");
                        maxCorner = new Point3d(Double.valueOf(maxCornerArray[0]),
                                Double.valueOf(maxCornerArray[1]),
                                Double.valueOf(maxCornerArray[2]));
                        break;
                    case "split":
                        String[] splitArray = value.split(" ");
                        split = new Point3i(Integer.valueOf(splitArray[0]),
                                Integer.valueOf(splitArray[1]),
                                Integer.valueOf(splitArray[2]));
                        break;
                    case "type":
                        if (value.equalsIgnoreCase("als")) {
                            type = Type.ALS;
                        } else if (value.equalsIgnoreCase("tls")) {
                            type = Type.TLS;
                        } else {
                            throw new IOException("Unsupported Lidar scanner type. Expected ALS/TLS. Found " + value);
                        }
                        break;
                    case "res":
                        resolution = Float.valueOf(value);
                        break;
                    case "max_pad":
                        maxPAD = Float.valueOf(value);
                        break;
                    case "lad_type":
                        if (value.contains("=")) {
                            // either two-beta parameters or ellipsoidal distribution
                            String[] split = value.split("=");
                            ladType = LeafAngleDistribution.Type.fromString(split[0]);
                            split[1] = split[1].replace("[", "");
                            split[1] = split[1].replace("]", "");
                            String[] params = split[1].split(";");
                            ladParams = new double[params.length];
                            for (int i = 0; i < params.length; i++) {
                                ladParams[i] = Double.valueOf(params[i]);
                            }
                        } else {
                            // other distribution without parameters
                            ladType = LeafAngleDistribution.Type.fromString(value);
                        }
                        break;
                    case "build-version":
                        buildVersion = value;
                        break;
                    case "fraction-digits":
                        nFractionDigits = Integer.valueOf(value);
                        break;
                }
            }

        } catch (IOException | NumberFormatException ex) {
            throw new IOException("Header is invalid", ex);
        }
    }

    public void readFromVoxelFile(File voxelFile) throws Exception {

        readMetadata(voxelFile);
        readColumnNames(voxelFile);
    }

    private void readMetadata(File voxelFile) throws Exception {

        try (BufferedReader reader = new BufferedReader(new FileReader(voxelFile))) {

            String identifier = reader.readLine();

            if (!identifier.equals("VOXEL SPACE")) {
                reader.close();
                throw new Exception("Voxel file is invalid, VOXEL SPACE identifier is missing");
            }

            properties.clear();
            nLine = 0;

            String header;
            while (null != (header = reader.readLine().trim())) {

                // increment line number
                nLine++;

                // ignore empty line
                if (!header.isEmpty()) {
                    // header is over, leave the loop
                    if (!header.startsWith("#")) {
                        break;
                    }

                    // valid header, extract parameters
                    // may be several parameters per line
                    String[] parameters = header.split("#");
                    // loop over the parameters
                    for (String parameter : parameters) {
                        if (parameter.isEmpty()) {
                            continue;
                        }
                        // expected parameter format key:value
                        String[] key_value = parameter.split(":");
                        if (key_value.length != 2) {
                            LOGGER.warn("Wrong header parameter format. Expected key:value. Found " + parameter);
                            continue;
                        }
                        String key = key_value[0].trim();
                        String value = key_value[1].trim();
                        properties.put(key, value);
                    }
                }
            }
            // increment to include column name in header
            nLine++;

            // initializes well known parameters
            presetProperties();

        } catch (FileNotFoundException ex) {
            throw new Exception("Cannot find voxel file", ex);
        } catch (IOException ex) {
            throw new Exception("Cannot read voxel file", ex);
        }
    }

    private void readColumnNames(File voxelFile) throws IOException {

        try (BufferedReader reader = new BufferedReader(new FileReader(voxelFile))) {

            // skip headers
            FileManager.skipLines(reader, nLine - 1);

            // read column name
            columnNames = reader.readLine().split(" ");
            columnNamesList = new ArrayList(columnNames.length);
            columnNamesList.addAll(Arrays.asList(columnNames));
        }
    }

    private void updateVoxelSize() {

        double boundingBoxSizeX = maxCorner.x - minCorner.x;
        double boundingBoxSizeY = maxCorner.y - minCorner.y;
        double boundingBoxSizeZ = maxCorner.z - minCorner.z;

        this.voxelSize.x = boundingBoxSizeX / split.x;
        this.voxelSize.y = boundingBoxSizeY / split.y;
        this.voxelSize.z = boundingBoxSizeZ / split.z;
    }

    public int getNLine() {
        return nLine;
    }

    public Point3d getMinCorner() {
        return minCorner;
    }

    public Point3d getMaxCorner() {
        return maxCorner;
    }

    public Point3i getSplit() {
        return split;
    }

    public Type getType() {
        return type;
    }

    public float getResolution() {
        return resolution;
    }

    public Point3d getVoxelSize() {

        if (voxelSize.x == 0 && voxelSize.y == 0 && voxelSize.z == 0) {
            updateVoxelSize();
        }
        return voxelSize;
    }

    public float getMaxPAD() {
        return maxPAD;
    }

    public String[] getColumnNames() {
        return columnNames;
    }

    public List<String> getColumnNamesList() {

        return columnNamesList;
    }

    public void setMinCorner(Point3d minCorner) {
        this.minCorner = minCorner;
    }

    public void setMaxCorner(Point3d maxCorner) {
        this.maxCorner = maxCorner;
    }

    public void setSplit(Point3i split) {
        this.split = split;
    }

    public void setType(Type type) {
        this.type = type;
        properties.put("type", type.toString());
    }

    public void setResolution(float resolution) {
        this.resolution = resolution;
        properties.put("res", String.valueOf(resolution));
    }

    public void setResolution(double resolution) {

        this.resolution = (float) resolution;
        properties.put("res", String.valueOf(this.resolution));

        if (minCorner != null && maxCorner != null) {

            split = new Point3i(
                    (int) Math.ceil((maxCorner.getX() - minCorner.getX()) / resolution),
                    (int) Math.ceil((maxCorner.getY() - minCorner.getY()) / resolution),
                    (int) Math.ceil((maxCorner.getZ() - minCorner.getZ()) / resolution));
        }
    }

    public void setMaxPAD(float maxPAD) {
        this.maxPAD = maxPAD;
        properties.put("max_pad", String.valueOf(maxPAD));
    }

    public void setColumnNames(String[] columnNames) {
        this.columnNames = columnNames;
    }

    public LeafAngleDistribution.Type getLadType() {
        return ladType;
    }

    public void setLadType(LeafAngleDistribution.Type ladType) {
        this.ladType = ladType;
        properties.put("lad_type", ladType.toString());
    }

    public double[] getLadParams() {
        return ladParams;
    }

    public void setLadParams(double[] ladParams) {
        this.ladParams = ladParams;
        if (ladType == LeafAngleDistribution.Type.TWO_PARAMETER_BETA || ladType == LeafAngleDistribution.Type.ELLIPSOIDAL) {
            String lad = ladType.toString() + "=[";
            for (int i = 0; i < ladParams.length; i++) {
                if (i != 0) {
                    lad += ";";
                }
                lad += ladParams[i];
            }
            lad += "]";
            properties.put("lad_type", lad);
        }
    }

    public String headerToString() {

        // update version
        properties.put("build-version", getVersion());

        StringBuilder metadata = new StringBuilder();
        for (String key : properties.keySet()) {
            // parameters min_corner, max_corner and split go on separate lines
            if (key.equalsIgnoreCase("min_corner")
                    || key.equalsIgnoreCase("max_corner")
                    || key.equalsIgnoreCase("split")) {
                continue;
            }
            String value = properties.get(key);
            metadata.append("#").append(key).append(":").append(value).append(" ");
        }

        String result = "VOXEL SPACE\n"
                + "#min_corner: " + minCorner.x + " " + minCorner.y + " " + minCorner.z + "\n"
                + "#max_corner: " + maxCorner.x + " " + maxCorner.y + " " + maxCorner.z + "\n"
                + "#split: " + split.x + " " + split.y + " " + split.z + "\n"
                + metadata.toString().trim();

        return result;
    }

    public String getVersion() {

        if (null == buildVersion) {
            try {
                buildVersion = ManifestMetaData.getValue(VoxelSpaceInfos.class, "Implementation-Version");
            } catch (Exception ex) {
                buildVersion = "UNDEF";
            }
        }
        return buildVersion;
    }

    public void setFractionDigits(int nFractionDigits) {
        this.nFractionDigits = nFractionDigits;
        properties.put("fraction-digits", String.valueOf(nFractionDigits));
    }

    public int getFractionDigits() {
        return nFractionDigits;
    }
}
