/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package fr.amap.lidar.amapvox.voxelisation.tls;

import fr.amap.commons.util.CallableTaskAdapter;
import fr.amap.commons.util.ProcessingAdapter;
import fr.amap.commons.util.filter.Filter;
import fr.amap.lidar.amapvox.commons.LidarScan;
import fr.amap.lidar.amapvox.shot.filter.EchoRankFilter;
import fr.amap.lidar.amapvox.util.AVoxTask;
import fr.amap.lidar.amapvox.util.FinalCounter;
import fr.amap.lidar.amapvox.voxelisation.configuration.TLSVoxCfg;
import fr.amap.lidar.amapvox.voxelisation.configuration.VoxMergingCfg;
import fr.amap.lidar.amapvox.voxelisation.postproc.VoxelFileMerging;
import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import org.apache.log4j.Logger;

/**
 *
 * @author pverley
 */
public class MultiScanTLSVoxelisation extends AVoxTask {

    private final Class voxelisationClass;

    private final static Logger LOGGER = Logger.getLogger(RSPVoxelisation.class);

    private final List<TLSVoxelisation> tasks;
    private VoxMergingCfg mergingCfg;

    public MultiScanTLSVoxelisation(File file, int ncpu, Class voxelisationClass) {
        super(file, ncpu);
        this.voxelisationClass = voxelisationClass;
        tasks = new ArrayList();
    }

    @Override
    public String getName() {
        return "TLS voxelisation";
    }

    @Override
    public void init() throws Exception {

        tasks.clear();
        TLSVoxCfg mainCfg = new TLSVoxCfg();
        mainCfg.readConfiguration(file);

        // list of output files
        List<File> outputFiles = new ArrayList();

        // loop over single TLS scans
        FinalCounter nDone = new FinalCounter(0);
        int nScan = mainCfg.getLidarScans().size();
        for (LidarScan scan : mainCfg.getLidarScans()) {

            TLSVoxCfg cfg = new TLSVoxCfg();
            cfg.readConfiguration(file);
            cfg.setInputFile(scan.file);
            File outputFile = new File(mainCfg.getOutputFile().getAbsolutePath() + "/" + scan.file.getName() + ".vox");
            outputFiles.add(outputFile);
            cfg.setOutputFile(outputFile);
            cfg.setSopMatrix(scan.matrix);

            // echo weight by file
            if (null != cfg.getVoxelParameters().getEchoesWeightByFileParams()) {
                cfg.getVoxelParameters().getEchoesWeightByFileParams().setScanName(scan.file.getName());
            }

            // echo filter by file
            for (Filter filter : cfg.getEchoFilters()) {
                if (filter instanceof EchoRankFilter) {
                    ((EchoRankFilter) filter).setScanName(scan.file.getName());
                    break;
                }
            }

            // add current scan voxelization to the list of configurations
            TLSVoxelisation voxelisation = (TLSVoxelisation) voxelisationClass.getConstructor(TLSVoxCfg.class, int.class).newInstance(cfg, 1);
            voxelisation.init();
            voxelisation.addCallableTaskListener(new CallableTaskAdapter() {
                @Override
                public void onSucceeded() {
                    nDone.increment();
                    String msg = "Voxelization, scan " + nDone.getValue() + "/" + nScan;
                    MultiScanTLSVoxelisation.this.fireProgress(msg, nDone.getValue(), nScan);
                }
            });
            tasks.add(voxelisation);
        }

        // Vox file merging
        if (mainCfg.getVoxelParameters().isMergingAfter()) {
            File mfile = mainCfg.getVoxelParameters().getMergedFile();
            mergingCfg = new VoxMergingCfg(mfile, mainCfg.getVoxelParameters(), outputFiles);
        }
    }

    @Override
    public File[] call() throws Exception {

        LOGGER.info("TLS Voxelization started");

        // list of output files
        ArrayList<File> outputFiles = new ArrayList();

        if (!tasks.isEmpty()) {
            try {
                // run RXP voxelisation tasks
                ExecutorService exec = Executors.newFixedThreadPool(Math.min(ncpu, tasks.size()));
                List<Future<File[]>> results = exec.invokeAll(tasks);
                for (Future<File[]> result : results) {
                    outputFiles.addAll(Arrays.asList(result.get()));
                }

                if (isCancelled()) {
                    // returns what has been done so far
                    return outputFiles.toArray(new File[outputFiles.size()]);
                }

                // vox file merging
                if (null != mergingCfg) {
                    VoxelFileMerging merging = new VoxelFileMerging(mergingCfg, 1);
                    merging.addProcessingListener(new ProcessingAdapter() {
                        @Override
                        public void processingStepProgress(String progressMsg, long progress, long max) {
                            fireProgress(progressMsg, progress, max);
                        }
                    });
                    merging.init();
                    outputFiles.addAll(Arrays.asList(merging.call()));
                }

            } catch (InterruptedException | NullPointerException ex) {
                fireCancelled();
                throw ex;
            }
        }

        return outputFiles.toArray(new File[outputFiles.size()]);
    }

    @Override
    public void setCancelled(boolean cancelled) {
        super.setCancelled(cancelled);
        // propagate cancellation to sub tasks
        for (TLSVoxelisation task : tasks) {
            task.setCancelled(cancelled);
        }
    }
}
